/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   INCLUDE:   finel.h   ***************
*/

/*  DESCRIPTION:
     global variables and definitions for GeoFEST finite element program
*/
/*   EOP   */
/*---------------------------------------------------------------------------*/
/*
***                           File finel.h                         ***
***                        GeoFEST version 4.3                    ***
*** Copyright (c) 2002, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
***/
/*
***GeoFEST version 4.3***
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining 
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
*/

/* global variables and definitions for finite element program */

#define REVNO "4.3"

/* finite element standard defs for 2-d  or 3-d code */

/* define system types                  */
#define  INTERIOR     0
#define  SHARED       1

/* definitions specific to this version */
#define MAXEL   5000
#define MAX_GROUP  10
#define MAX_PROCS 32
#define MAX_NEIGHBOR 5
/* generous a priori bound on how many nodes connect to a node, on average */
/* to be used to size adjacency array icol(number of nonzeros). */
#define MAX_CONNECT 50

/* macro definitions */

#define GLOBALIZE \
        ((solve->neq_all == fe_sys.neq) ? (GLOBALIZE_F) : (GLOBALIZE_S))
#define ELASTIC   0
#define VISCO     1
#define FACBACK   0
#define BACK      1
#define ITER      2
#define SUB       0
#define FULL      1
#define DIRECT      1
#define PCG      2
#define BILIN         1
#define SERENDIP      2
#define QUAKE         3
#define TET           4
#define TRUSS         5
#define TRILIN        6
#define GENERATE     1
#define OUTPUT       2
#define SHAPE        3
#define FORMS_ELAS   4
#define FORMS_STEP   5
#define RHS_ELAS     6
#define RHS_STEP     7
#define BC_ELAS      8
#define BC_STEP      9
#define COL_HT      10
#define E_STRESS    11
#define V_STRESS    12
#define DUMP        13
#define RESTORE     14
#define NEW_LM      15
#define FAIL_CHECK  16
#define SMOOTH_BEGIN    17
#define SMOOTH_END      18
#define REORDER         19
#define ESPROD         20
#define SETUP            1
#define DRYRUN           2
#define GLOBALIZE_S      3
#define GLOBALIZE_F      4
#define DOT_PROD         5
#define FOLD2            6
#define FOLD4            7
#define FOLD6            8
#define FOLD8            9
#define FOLDI            10
#define CGTOL      1.0e-05


/*  global parameters  */

EXTERN  int             attempt ;

/******************************************/  
 struct system_info
    {
     int               numnp ;
     int                 nsd ;
     int                 neq ;
     int                ndof ;
     int              nrates ;
     int         save_shape  ;
     int         solver     ;
    } ;
typedef struct system_info SYSTEM_INFO ;
EXTERN  SYSTEM_INFO   fe_sys ;

/******************************************/  
 struct  time_info
      {
       int   ntime_grp             ;
       int   *nsteps               ;
       real *alpha                ;  
       real *delt                 ;
       int   nreform               ;
       int   nprints               ;
       real *prt_time             ;
       int   nprt_nodes            ;
       int   *node_list            ;
       int   nprt_elem             ;
       int   *elem_list            ;
       int   restart               ;
       int   sav_state             ;
       char  startfile[80]         ;
       char  savefile[80]          ;
       int    quake_prt            ;
       int    fail                 ;
       int    elastic              ;
       real  alpha_delt           ;
       real  dt                   ;
       real  time                 ;
       real  slip_interval        ;
       real  next_slip            ;
       int   do_slip              ;
       int    paused               ;
       int    nbackup              ;
       int  nsub ;
       real *maxstep ;
       real *endtime ;
       real max_strain ;
       real scale_ratio ;
      } ;
typedef   struct time_info   TIME_INFO  ;
EXTERN  TIME_INFO    time_data  ;
 
/******************************************/  
 struct global_data
    {
     int       n_group ;
     real     *coords ;
     real     *displ  ;
     real     *del_displ ;
     real     *forv ;
     real     *rate ;
     real     *bctime ;
     real     *pnode ;        /* nodal pressures for smoothing */
     real     *yp ;       /* diagonalized matrix for smoothing */
     int     *g_nodes ;
     int     *id_pointer ;
     int     *eqtype ;
     int     *permut ; /* profile-optimizing permutation of global nodes */
    } ;
typedef struct global_data GLOBAL_DATA ;
EXTERN  GLOBAL_DATA  global ;

/******************************************/  
 struct element_data
    {
     int     nel    ;
     real     *stiff ;
     real     *shape ;
     real     *deter   ;
     real     *stress  ;
     real     *bta     ;
     real     *dbar    ;
     real     *dilat   ;
     int      *is_split ;
     int     mat    ;
     int     is_bc  ;
     int     degen  ;
     int     nint  ;
     int     *ien   ;
     int     *lm    ;
    } ;
typedef struct element_data ELEMENT_DATA ;

/******************************************/  
 struct element_info
    {
     int      type        ;
     int      numel       ;
     int      numat       ;
     int      numsuf      ;
     int      nfnode      ;
     int      nen         ;
     int      nee         ;
     int      el_size     ;
     int      stress_size ;
     int      dbar_size   ;
     int      failed      ;
     int    *g_elem     ;
     int    *surf_list  ;
     real    *surf_trac  ;
     int    *slip_list  ;
     real    *slip_val   ;
    } ;
typedef struct element_info ELEMENT_INFO ;

/******************************************/  
 struct element_mat
    {
     real     *mu     ;
     real     *lambda ;
     real     *visc   ;
     real     *bforce ;
     int     *plastic ;
    } ;
typedef struct element_mat ELEMENT_MAT ;

/******************************************/  
 struct group_dir
    {
     ELEMENT_DATA     *el_data ;
     ELEMENT_INFO     *el_info ;
     ELEMENT_MAT      *el_mat  ;
     int              group_num ;
    } ;
typedef struct group_dir GROUP ;
EXTERN  GROUP  groups[MAX_GROUP] ;

/******************************************/  
 struct rhs_data
    {
     real     *full_rhs        ;
     real     *incr_rhs        ;
     real     *last_result     ;
    } ;
typedef struct rhs_data RHS_DATA ;
EXTERN  RHS_DATA  force ;

/******************************************/  
 struct profile_data
    {
     real     *matrix ;
     int       *diag   ;
     int       neq     ;
     int       size    ;
    } ;
typedef struct profile_data PROFILE ;
EXTERN  PROFILE  stiff ;

 struct cg_data
    {
     int       neq     ;
     real     *precond ;
     real     *r ;
     real     *d ;
     real     *temp ;
     real     *hist ;
     real     norm ;
    } ;
typedef struct cg_data CONGRAD ;
EXTERN  CONGRAD  pcg ;

/******************************************/  
typedef real *COLUMN ;
COLUMN  *matcol ;
/******************************************/  

EXTERN  FILE  *in_file  ;
EXTERN  FILE  *out_file ;
EXTERN  FILE  *dump_file ;
EXTERN  FILE  *cgconv_file ;
EXTERN  int  bugflag ;
