/*
***                           File generat.c                      ***
***                        GeoFEST version 4.3                    ***
*** Copyright (c) 2002, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains modules for generating and reading 
*** node and element data and related isoparametric tasks 
***
***          == gen_element ==
***          == next_element ==
***          == read_surf ==
***          == read_slip ==
***          == splitfind ==
***          == load_element ==
***          == gen_map ==
***          == gen_real ==
***          == gen_isopar_grid ==
***          == gen_shape_grid ==
***          == dot_sh ==
***
***/

/*
***GeoFEST version 4.3***
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining 
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
 $Log: generat.c,v $
 Revision 1.3  2002/08/23 18:46:49  jwp
 Restored keyword strings, Log and REVNO festures.
 Makefile now has .h dependencies.

 Revision 1.1  2002/07/12 19:05:14  jwp
 Initial revision

*/

#include <stdio.h>
#include <math.h>
#include "mydefs.h"
#include "errcodes.h"
#define EXTERN extern
#include "finel.h"

/* GC: global coordinate value of co (0,1,2 => x,y,z) for node nd */
#define GC(co,nd) (global.coords[(co) + (nd)*nsd])

/* XV:  tet basis vector, component co (0,1,2) for tet node en from node 0*/
#define XV(co,en) (GC(co,el_pt->ien[(en)] - 1) - GC(co,el_pt->ien[0] - 1))

static
     int      el_inc[3] ,
              nod_inc[3] ,
              el_end[3] ;
static
     int         current_element  ,
                 node             ,
                 gen_order        ,
                 nodes_gen[3]     ,
                 node_inc[3]      ;
static int    last_node ;
static int      *temp_list , *temp_where ;
static real     *temp_trac , *temp_amount ;
static
     real  temp[3][8]   ;

static
     real shape[8] = { ZERO , ZERO , ZERO , ZERO
              , ZERO , ZERO , ZERO , ZERO } ;


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_element   ***************
*/

/*  INTERFACE:   */

    gen_element(
                GROUP   *grp_ptr    /* pointer to current element group */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_element
** gen_element generates element storage and ien array
** It reads in element group information, individual elements,
** and surface traction records.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    char CVS_Id[] = "@(#) CVS id $Id: generat.c,v 1.3 2002/08/23 18:46:49 jwp Exp $, Revision: " REVNO;
    COUNTER  i , j , k , m , nel ;
    int       type , numsuf , ndum , ndof , nsd , nsf , plastic ,
              tally , nfnode , count , this_node ;
    int       nel_local , mat , mine , numel , numat , nen , nee ;
    int       stress_size  , dbar_size , nint ; 
    int       x_index , y_index , x_offset , y_offset , next_el ;
    int       *fnode ;
    real     *fvec ;
    real     coeff , exponent , slip , ivec[3] , svec[3] , cross[3] ;
    real     xctr , yctr , zctr , dot , mult , norm ;
    ELEMENT_DATA   *el_pt ;
    int       ien_current[8]  ;
    int       ien_temp[8];

/*  begin with element constants . . .  */

     fscanf(in_file,"%d",&numel) ;

     fscanf(in_file,"%d%d%d%d%d",&type,&numat,&numsuf,&ndum,&nfnode) ;
     (grp_ptr->el_info)->numel = numel ;
     (grp_ptr->el_info)->type = type ;
     (grp_ptr->el_info)->numat = numat ;
     (grp_ptr->el_info)->numsuf = numsuf ;
     (grp_ptr->el_info)->nfnode = nfnode ;
     (grp_ptr->el_info)->failed = 0 ;

     grp_ptr->el_data =
        (ELEMENT_DATA *) calloc(numel,sizeof(ELEMENT_DATA)) ;
     if (grp_ptr->el_data == NULL) attempt = ELD_MEM ;

     completion() ;

     switch( type )
        {
         case BILIN:
         case TET:
            nen = (grp_ptr->el_info)->nen = 4 ;
            break ;

         case SERENDIP:
         case TRILIN:
            nen = (grp_ptr->el_info)->nen = 8 ;
            break ;
        }
        
     ndof = fe_sys.ndof ;
     nsd = fe_sys.nsd ;
     nee = (grp_ptr->el_info)->nee = nen*ndof ;
     (grp_ptr->el_info)->el_size = nee*nee ;
     if( type == BILIN || type == SERENDIP || type == TET || type == TRILIN )
        {
         switch( ndof )
            {
             case 1:
                stress_size = (grp_ptr->el_info)->stress_size = 2 ;
                dbar_size   = (grp_ptr->el_info)->dbar_size   = 3 ;
                break ;
             case 2:
                stress_size = (grp_ptr->el_info)->stress_size = 4 ;
                dbar_size   = (grp_ptr->el_info)->dbar_size   = 10 ;
                break ;
             case 3:
                stress_size = (grp_ptr->el_info)->stress_size = 6 ;
                dbar_size   = (grp_ptr->el_info)->dbar_size   = 21 ;
                break ;
            }
         nsf = ndof ;
        }
     
/*  next get element material properties . . .  */

     (grp_ptr->el_mat)->mu     =
              (real *) calloc(numat,sizeof(real)) ;
     (grp_ptr->el_mat)->lambda =
              (real *) calloc(numat,sizeof(real)) ;
     (grp_ptr->el_mat)->visc   =
            (real *) calloc(2*numat,sizeof(real)) ;
     (grp_ptr->el_mat)->bforce =
         (real *) calloc(ndof*numat,sizeof(real)) ;
     (grp_ptr->el_mat)->plastic=
              (int *) calloc(numat,sizeof(int)) ;
     
     for(i=0 ; i<numat ; i++)
        {
         fscanf(in_file,"%lf",(grp_ptr->el_mat)->mu+i) ;
         fscanf(in_file,"%lf",(grp_ptr->el_mat)->lambda+i) ;
         fscanf(in_file,"%lf%lf",&coeff,&exponent) ;
         *((grp_ptr->el_mat)->visc+2*i) = coeff ;
         *((grp_ptr->el_mat)->visc+2*i+1) = exponent ;
         if(coeff == ZERO || exponent == ZERO)
            *((grp_ptr->el_mat)->plastic+i) = 0 ;
         else if(exponent == ONE)
            *((grp_ptr->el_mat)->plastic+i) = 1 ;
         else
            *((grp_ptr->el_mat)->plastic+i) = 2 ;
            
/* printf("mat=%d  mu=%g  lambda=%g  coeff=%g  exp=%g\n",
           i+1,*((grp_ptr->el_mat)->mu+i),*((grp_ptr->el_mat)->lambda+i),
           coeff,exponent) ;

   printf("material #%d plasticity code = %d\n",
             i+1,*((grp_ptr->el_mat)->plastic+i) ) ;
*/
         for(j=0;j<ndof;j++)
            fscanf(in_file,"%lf",(grp_ptr->el_mat)->bforce+ndof*i+j) ;
        }
    
     for( nel=0 ; nel<numel ; nel++ )
        {
         el_pt = grp_ptr->el_data + nel ;
          
   /*  now allocate memory for the ien and lm arrays... */

         el_pt->ien = (int *) calloc(nen,sizeof(int)) ;
         if (el_pt->ien == NULL) attempt = IEN_MEM ;

         el_pt->lm = (int *) calloc(nee,sizeof(int)) ;
         if (el_pt->lm == NULL) attempt = LM_MEM ;

        }

        completion() ;

/*  now generate elements, filling up the arrays just created... */

      nel_local = 0 ;  /* zero-based element number */
      while (next_element( &current_element , &gen_order ) !=  0 )
           {
             fscanf(in_file,"%d",&mat);  
             mine = TRUE ;
             for ( i = 0 ; i < nen ; ++i)
                 fscanf(in_file,"%d",&ien_current[i]);  
             if ( gen_order != 0 )
                {
                 for ( i = 0 ; i < gen_order ; i++ )
                     fscanf(in_file,"%d",&el_inc[i]) ;  
                 for ( i = 0 ; i < gen_order ; i++ )
                     fscanf(in_file,"%d",&nod_inc[i]) ;  
                 for ( i = 0 ; i < gen_order ; i++ )
                     fscanf(in_file,"%d",&el_end[i]) ;  
                }
            if(mine)
               {
                if(gen_order == 0)
                   {
                    el_pt = grp_ptr->el_data + current_element - 1 ;
                    load_element(el_pt,grp_ptr->el_info,
                                 current_element,ien_current,mat) ;
                    nel_local++ ;
                   }

                if(gen_order == 1)
                   {
                    x_index = (-1) ;
                    next_el = current_element - el_inc[0] ;
                    while( next_el < el_end[0] )
                       {
                        x_index++ ;
                        next_el += el_inc[0] ;
                        for(i=0 ; i<nen ; i++)
                        ien_temp[i] =
                           ien_current[i] + nod_inc[0]*x_index ;
                        el_pt = grp_ptr->el_data + next_el - 1 ;
                        load_element(el_pt,grp_ptr->el_info,next_el,
                                     ien_temp,mat) ;
                        nel_local++ ;
    
                       }
                   }
    
                if(gen_order == 2)
                   {
                    y_offset = -el_inc[1] ;
                    y_index = (-1) ;
                    while( y_offset < el_end[1]-current_element )
                       {
                        y_offset += el_inc[1] ;
                        y_index++ ;
    
                        x_offset = -el_inc[0] ;
                        x_index = (-1) ;
                        while( x_offset < el_end[0]-current_element )
                           {
                            x_offset += el_inc[0] ;
                            x_index++ ;
                            next_el =
                               current_element + y_offset + x_offset ;
                            for(i=0 ; i<nen ; i++)
                              ien_temp[i] = ien_current[i] +
                                       nod_inc[0]*x_index +
                                       nod_inc[1]*y_index ;
                            el_pt = grp_ptr->el_data + next_el - 1 ;
                            load_element(el_pt,grp_ptr->el_info,next_el,
                                         ien_temp,mat) ;
                            nel_local++ ;
                           }
                       }
                   }
               }
            completion() ;
           }

      if(numel != nel_local) 
         {
          attempt = NEL_ERR ;
         }
      completion() ;

      printf("Elements all read in . . .\n") ;

     for( nel=0 ; nel<numel ; nel++ )
        {
         el_pt = grp_ptr->el_data + nel ;
         m = el_pt->mat ;
         plastic = *( (grp_ptr->el_mat) -> plastic + m ) ;
         nint = el_pt->nint ;
          
   /*  next allocate memory for the stresses and visco arrays... */

         
         if( type == BILIN || type == SERENDIP || type == TET || type == TRILIN)
            {
             el_pt->stress =
                ( real * ) calloc( nint*stress_size , sizeof(real)) ;
             if (el_pt->stress == NULL) attempt = STRESS_MEM  ;
             
             el_pt->dilat =
                ( real * ) calloc( nint , sizeof(real)) ;
             if (el_pt->dilat == NULL) attempt = STRESS_MEM  ;

             el_pt->is_split =
                ( int * ) calloc( 8 , sizeof(int)) ;
  /* there should never be more than 7 split nodes in an element, I hope */              
             if (el_pt->is_split == NULL) attempt = STRESS_MEM  ;
             *(el_pt->is_split) = 0 ; /* initialize just to be sure */

             if( plastic )
                {
                 el_pt->bta =
                 ( real * ) calloc( nint*stress_size , sizeof(real)) ;
                 if (el_pt->bta == NULL) attempt = BTA_MEM  ;
            
                 el_pt->dbar =
                 ( real * ) calloc( nint*dbar_size , sizeof(real)) ;
                 if (el_pt->dbar == NULL) attempt = DBAR_MEM  ;

                }
            }
   /*  if PCG is being used, allocate memory for element stiffness storage */
         if( fe_sys.solver == PCG )
            {
             el_pt->stiff =
                ( real * ) calloc( (grp_ptr->el_info)->el_size , sizeof(real)) ;
             if (el_pt->stiff == NULL) attempt = ELSTIFF_MEM  ;
            }
           
        }
      completion() ;

   /*  now allocate memory and read in surface tractions... */

      if( numsuf != 0 )
         {
          temp_list = (int *) calloc(numsuf,2*sizeof(int)) ;
          temp_trac = (real *) calloc(numsuf,nsf*sizeof(real)) ;
          
          tally = 0 ;
          
          read_surf( &tally , grp_ptr->el_data , temp_list , temp_trac ,
                     type , numel , ndof ) ;
                     
              (grp_ptr->el_info)->surf_list = temp_list ;
              (grp_ptr->el_info)->surf_trac = temp_trac ;
          printf("Surface tractions all read in . . .\n") ;
         }
      else
        printf("No surface tractions to process . . .\n") ;

   /*  now allocate memory for and read in and process split node faults... */

      if( nfnode == 0 )
         {
          printf("\nNo fault nodes; generation of element group #%d complete.\n",
             grp_ptr->group_num) ;
          return ;
         }
      fnode = (int *) calloc(nfnode,sizeof(int)) ;
      fvec = (real *) calloc(nfnode,7*sizeof(real)) ;
          
      read_slip( fnode , fvec , type , nfnode , ndof ) ;
                     
      temp_where = (int *) calloc(nfnode,2*sizeof(int)) ;
      temp_amount = (real *) calloc(nfnode,ndof*sizeof(real)) ;
      if (temp_where == NULL || temp_amount == NULL) attempt = SPLIT_MEM  ;
      completion() ;
      
      tally = 0 ;
      for( nel=0 ; nel<numel ; nel++ )
         {
          el_pt = grp_ptr->el_data + nel ;
          for(i=0 ; i<nen ; i++)
             {
              this_node = *(el_pt->ien + i) ;
                  j = splitfind(this_node,fnode,nfnode) ;
/* j is the zero-based index number of the searched-for split node input entry */

                  if(j != -1)
                     {
                      tally++ ;
					  temp_where = (int *) realloc(temp_where,2*sizeof(int)*tally) ;
					  temp_amount = (real *) realloc(temp_amount,ndof*sizeof(real)*tally) ;
					  if (temp_where == NULL || temp_amount == NULL) attempt = SPLIT_MEM  ;
                      
                      *(temp_where+2*(tally-1)) = nel ;         /* zero-based el# */
                      *(temp_where+2*(tally-1)+1) = this_node ; /* one-based node# */
                      
                      count = (*(el_pt->is_split)) + 1 ;
                      *(el_pt->is_split) = count ;
                      *(el_pt->is_split + count) = tally-1 ;
 /* flag and indices set for easy lookup of split nodes later during stress calc */

                      ivec[0] = *(fvec + 7*j) ;
                      ivec[1] = *(fvec + 7*j + 1) ; /* ivec is the direction perp to slip */
                      ivec[2] = *(fvec + 7*j + 2) ;
                      svec[0] = *(fvec + 7*j + 3) ;
                      svec[1] = *(fvec + 7*j + 4) ; /* svec is the slip vector direction */
                      svec[2] = *(fvec + 7*j + 5) ;
                      slip    = *(fvec + 7*j + 6) ; /* slip is the full slip amplitude */
                     
                      norm = sqrt(ivec[0]*ivec[0] + ivec[1]*ivec[1] + ivec[2]*ivec[2]) ;
                      if(fabs(norm - ONE) > 1.0e-03)
                         {
                          printf("ivec not normalized!\n");
                          printf("node #%d , splitnode #%d , instance #%d\n",
                                 this_node,j+1,tally) ;
                          printf("ivec = [ %g %g %g ]\n",ivec[0],ivec[1],ivec[2]) ;
                          printf("|ivec| = %g\n", norm );
                          attempt = IVEC_NORM ;
                         }
                      norm = sqrt(svec[0]*svec[0] + svec[1]*svec[1] + svec[2]*svec[2]) ;
                      if(fabs(norm - ONE) > 1.0e-03)
                         {
                          printf("svec not normalized!\n");
                          printf("node #%d , splitnode #%d , instance #%d\n",
                                 this_node,j+1,tally) ;
                          printf("svec = [ %g %g %g ]\n",svec[0],svec[1],svec[2]) ;
                          printf("|svec| = %g\n", norm );
                          attempt = SVEC_NORM ;
                         }
                      if(fabs(ivec[0]*svec[0]+ivec[1]*svec[1]+ivec[2]*svec[2]) > 1.0e-03)
                         {
                          printf("ivec and svec not perpendicular!\n");
                          attempt = ISVEC_PERP ;
                         }
                     
                      cross[0] = svec[1]*ivec[2] - svec[2]*ivec[1] ;
                      cross[1] = svec[2]*ivec[0] - svec[0]*ivec[2] ;
                      cross[2] = svec[0]*ivec[1] - svec[1]*ivec[0] ;
                      
                      xctr = ZERO ;
                      yctr = ZERO ;
                      zctr = ZERO ;
                      for(k=0 ; k<nen ; k++)
                         {
                          xctr += *(global.coords + nsd * ( *(el_pt->ien + k) - 1 ) ) ;
                          yctr += *(global.coords + nsd * ( *(el_pt->ien + k) - 1 ) + 1 ) ;
                          if(nsd == 3)
                          zctr += *(global.coords + nsd * ( *(el_pt->ien + k) - 1 ) + 2 ) ;
                         }
                      xctr = xctr/(real)nen - *(global.coords + nsd * ( this_node - 1 ) ) ;
                      yctr = yctr/(real)nen - *(global.coords + nsd * ( this_node - 1 ) + 1 ) ;
                      if(nsd == 3)
                      zctr = zctr/(real)nen - *(global.coords + nsd * ( this_node - 1 ) + 2 ) ;
                      
                      dot = cross[0]*xctr + cross[1]*yctr + cross[2]*zctr ;
                      if( dot > ZERO )
                         mult = slip * PT5 ;
                      else
                         mult = -slip * PT5 ;
                      switch( ndof )
                         {
                          case 1:
                              *(temp_amount+ndof*(tally-1)) = svec[2]*mult ;
                              break ;
                          case 2:
                              *(temp_amount+ndof*(tally-1)) = svec[0]*mult ;
                              *(temp_amount+ndof*(tally-1)+1) = svec[1]*mult ;
                              break ;
                          case 3:
                              *(temp_amount+ndof*(tally-1)) = svec[0]*mult ;
                              *(temp_amount+ndof*(tally-1)+1) = svec[1]*mult ;
                              *(temp_amount+ndof*(tally-1)+2) = svec[2]*mult ;
                              break ;
                         }

                     }
                  
             }     /* loop i over element nodes */
          completion() ;
         }         /* loop nel over elements */
         

      (grp_ptr->el_info)->nfnode = tally ; /* nfnode now contains the total number of instances! */
      (grp_ptr->el_info)->slip_list = temp_where ;
      (grp_ptr->el_info)->slip_val = temp_amount ;
      free( fnode ) ;
      free( fvec ) ;
      printf("Split nodes all processed . . .\n") ;


      printf("\nGeneration of element group #%d complete.\n",
             grp_ptr->group_num) ;
      fflush(stdout) ;
   }

/************************** end of gen_element *******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   next_element   ***************
*/

/*  INTERFACE:   */

   int  next_element(
                     int   *node_pointer , /* pointer to hold node */
                     int   *gen_pointer   /* pointer to gen parameter */
                    )

/*  RETURN VALUE:  number of the next node or object read in */
/*  DESCRIPTION:   */
/*
** Function next_element
** next_element reads the first part of a record in a double 
** null-terminated list, so the calling function may 
** catch the termination.
**
** The use of this function is not limited to finite elements, but is
** also employed for other lists, such as nodes and tractions.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

            {
             fscanf(in_file,"%d",node_pointer) ;  
             fscanf(in_file,"%d",gen_pointer) ;

             return(*node_pointer);
            }

/************************** end of next_element *******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   read_surf   ***************
*/

/*  INTERFACE:   */

     read_surf(
               int  *tally ,  /* counter for records read in */
               ELEMENT_DATA  *el_pt ,  /* ptr to current element struct */
               int  *temp_list ,  /* storage for surf location data */
               real  *temp_trac ,  /* storage for surf traction data */
               int  type ,    /* element type number */
               int  numel ,  /* number of elements */
               int  ndof    /* number of disp degrees of freedom */
              )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine read_surf
** read_surf reads in surface traction records.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int  nel , side , grp , n , i , j ;
    ELEMENT_DATA  *wall_ptr ;
    
    switch( type )
       {
        case BILIN:
        case SERENDIP:
        case TET:
        case TRILIN:
          while (next_element( &nel , &side ) !=  0 )
             {
              n = nel - 1 ;

              *(temp_list+2*(*tally)) = n ;
              *(temp_list+2*(*tally)+1) = side ;
              for(j=0;j<ndof;j++)
                fscanf(in_file,"%lf",temp_trac+ndof*(*tally)+j) ;
              (*tally)++ ;
             }
          break ;

        case QUAKE:
          break ;
       }
   }
/************************** end of read_surf *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   read_slip   ***************
*/

/*  INTERFACE:   */

     read_slip(
               int  *fnode ,  /* global numbers of split nodes */
               real  *fvec ,  /* slip and plane vectors */
               int  type ,    /* element type number */
               int  nfnode ,  /* number of split nodes */
               int  ndof    /* number of disp degrees of freedom */
              )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine read_slip
** read_slip reads split-node records for symmetric fault 
** slip accumulation.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int  nel , n , j ;
    
    switch( type )
       {
        case BILIN:
        case SERENDIP:
        case TET:
        case TRILIN:
          for(n=0 ; n<nfnode ; n++)
             {
              fscanf(in_file,"%d%lf%lf%lf%lf%lf%lf%lf",
                fnode+n , fvec+7*n , fvec+7*n+1 , fvec+7*n+2 ,
                 fvec+7*n+3 , fvec+7*n+4 , fvec+7*n+5 , fvec+7*n+6 ) ;
  /* node # , ivec[0] , ivec[1] , ivec[2] , svec[0] , svec[1] , svec[2] , slip */

             }
          break ;

        case QUAKE:
          break ;
       }
   }
/************************** end of read_slip *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   splitfind   ***************
*/

/*  INTERFACE:   */

 int splitfind(
               int   qnode ,  /* query node to search for */
               int  *fnode ,  /* global numbers of split nodes */
               int  nfnode    /* number of split nodes */
              )

/*  RETURN VALUE:   index number of found split node entry; -1 if not found */
/*  DESCRIPTION:   */
/*
** function splitfind
** splitfind performs an efficient bisection search on the list of split nodes 
** to see if qnode is among them.  If so, it returns the index number of
** that entry.  Otherwise, returns -1.  Assumes that the fnode list is
** provided sorted in ascending node number order!
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
 int  bot , mid , top ;
 
 bot = 0 ;
 top = nfnode-1 ;
 mid = top/2 ;
 
 if( !BETWEEN(fnode[bot],qnode,fnode[top]) )  return( -1 ) ;
 if(qnode == fnode[bot])  return( bot ) ;
 if(qnode == fnode[top])  return( top ) ;
 if(qnode == fnode[mid])  return( mid ) ;
 while((top-bot) > 2)
    {
     if( BETWEEN(fnode[bot],qnode,fnode[mid]) )
        {
         top = mid ;
         mid = bot + (top-bot)/2 ;
         if(qnode == fnode[mid])  return( mid ) ;
        }
     else
        {
         bot = mid ;
         mid = bot + (top-bot)/2 ;
         if(qnode == fnode[mid])  return( mid ) ;
        }
    }
 return( -1 ) ;
}
/************************** end of splitfind *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   load_element   ***************
*/

/*  INTERFACE:   */

  load_element(
               ELEMENT_DATA  *el_pt  ,  /* ptr to current element struct */
               ELEMENT_INFO  *info ,   /* ptr to element info struct */
               int      nel_glob ,      /* one-based global element number */
               int      ien[] ,     /* holding for ien assignments */
               int      mat       /* one-based material index number */
              )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine load_element 
** load_element loads element arrays with ien and equation data.
** The information for the element has been previously read into a 
** temporary array or generated. 
**
** This function checks for degeneracies of 2D elements,
** right-hand rule order of tet nodes (swapping when necessary),
** and fills in the ien and lm arrays for the element.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     COUNTER i , j , k , index , ieq , ndof , nsd ;
     int   node_number ;
     real s;
     int itmp;
     
     ndof = fe_sys.ndof ; 
     nsd = fe_sys.nsd ; 
     el_pt->nel =  nel_glob ;
     el_pt->mat = mat-1 ;        /* mat is stored in zero-based form  */

/* load the ien array with one-based node numbers ...  */

     for(i=0 ; i<info->nen ; i++)
        {
         if( (*(el_pt->ien + i) = ien[i]) )
	      {
           continue ;
	      }
         else
	      {
           attempt = IEN_ERR*1000 + nel_glob ;
	      }
             /* Error if glob_to_local returns zero local node # */
        }
/* jwp: if tet, check triple product and reorder if necessary. TP should
be > 0 if tet is right-handed. Check for determinant >0 (later) at all
wt points .  
*/

     if(info->nen == 4 && nsd == 3) {
        s = XV(0,1)*XV(1,2)*XV(2,3) - XV(2,1)*XV(1,2)*XV(0,3) +
            XV(1,1)*XV(2,2)*XV(0,3) - XV(0,1)*XV(2,2)*XV(1,3) +
            XV(2,1)*XV(0,2)*XV(1,3) - XV(1,1)*XV(0,2)*XV(2,3);

        if(s < 0) 
        {
           itmp          = el_pt->ien[3];
           el_pt->ien[3] = el_pt->ien[2];
           el_pt->ien[2] = itmp;
        }
     }

     el_pt->degen = 0 ;
     switch( info->type )
        {
         case BILIN:
             el_pt->nint = 4 ;
             if( ien[2] == ien[3] )
                el_pt->degen = 1 ;
             break ;

         case TET:
             el_pt->nint = 4 ;
             break ;

         case TRILIN:
             el_pt->nint = 6 ;
             if( ien[2] == ien[3] && ien[6] == ien[7] )
                el_pt->degen = 1 ; /* triangular prism */
             if( ien[2] == ien[3] && ien[4] == ien[5] && ien[4] == ien[6] && ien[4] == ien[7] )
                el_pt->degen = 2 ; /* tetrahedron */
             if(el_pt->degen != 0)  el_pt->nint = 8 ;
             break ;

         case SERENDIP:
             el_pt->nint = 9 ;
             if( ien[2] == ien[3] && ien[2] == ien[6] )
                el_pt->degen = 1 ;
             break ;
        }
/* and load the lm array with zero-based equation numbers  */

     el_pt->is_bc = 0 ;
     index = 0 ;
     for( j = 0 ; j < info->nen ; j++)
        {
         node_number = *(el_pt->ien + j) ;
         for( k = 0 ; k < ndof ; k++)
             {
              ieq = *(global.id_pointer + ndof*(node_number-1) + k) ;
              *(el_pt->lm + index) = ieq ;
              if( ieq == (-1) )
                 el_pt->is_bc = 1 ;
              index++ ;
             }
        }
    }

/************************** end of load_element ******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_map   ***************
*/

/*  INTERFACE:   */

    gen_map()

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_map
** gen_map fills the nodal activity map array.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    COUNTER  i , k ;
    int     bctype[3] , ndof ;
    int       current_node , gen_inc , end_node , next_node ,
              mine , x_inc , y_inc , end_x , end_y , x_offset ,
              y_offset ;

    ndof = fe_sys.ndof ; 
    last_node = 0 ;
    
     
    /* the map array contains space for 
    up to 3 d.o.f. type designators */

    printf("Start reading boundary info . . .\n") ;
     
    while (next_element( &current_node , &gen_order ) !=  0 )
      {
        for(k=0 ; k<ndof ; k++)
           fscanf(in_file,"%d",&bctype[k]) ;

        mine = TRUE ;
        if( mine )
           {
            last_node++ ;
            if(last_node > fe_sys.numnp)
               {
                attempt = NODES_ERR1 ;
                continue ;
               }
            
            for(k=0 ; k<ndof ; k++)
                *(global.eqtype+ndof*(current_node-1)+k) = bctype[k] ;
           }

    if(gen_order == 1)
       {
        fscanf(in_file,"%d%d",&gen_inc,&end_node);
        next_node = current_node ;
        while( next_node < end_node )
           {
            next_node += gen_inc ;
            if( mine )
               {
                last_node++ ;
                if(last_node > fe_sys.numnp)
                   {
                    attempt = NODES_ERR1 ;
                    continue ;
                   }
            
                for(k=0 ; k<ndof ; k++)
                    *(global.eqtype+ndof*(next_node-1)+k) = bctype[k] ;
               }
           }
       }
    
    if(gen_order == 2)
       {
        fscanf(in_file,"%d%d%d%d",&x_inc,&y_inc,&end_x,&end_y);
        y_offset = -y_inc ;
        while( y_offset < end_y-current_node )
           {
            y_offset += y_inc ;

            x_offset = -x_inc ;
            while( x_offset < end_x-current_node )
               {
                x_offset += x_inc ;
                next_node = current_node + y_offset + x_offset ;
                if( next_node != current_node )
                   {
                    if( mine )
                       {
                        last_node++ ;
                        if(last_node > fe_sys.numnp)
                           {
                            attempt = NODES_ERR1 ;
                            continue ;
                           }
            
                        for(k=0 ; k<ndof ; k++)
                            *(global.eqtype+ndof*(next_node-1)+k) = bctype[k] ;
                       }
                   }
               }
           }
       }
      }  /* end while */
     
     completion() ; 

     if(fe_sys.numnp != last_node)
       {
        attempt = NODES_ERR2 ;
        printf("Actual numnp = %d\n", last_node);
       }
     else
       {
        printf("Correct node tally.\n") ;
       }
     fflush(stdout) ;
     completion() ;
     
    }

/************************** end of gen_map ***********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_real   ***************
*/

/*  INTERFACE:   */

    gen_real(
             real      *array ,  /* array of generated values */
             int        dim      /* dimensionality of grid */
            )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_real
** gen_real reads or generates floating point global array.
** Used to read or generate nodes, displacement bcs, or 
** velocity bcs.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {

      COUNTER i , j  ;
      int    l_node ;
      real   value ;

      while (next_element( &node , &gen_order) != 0 )
            {
             if(gen_order == 0 )
               {
                l_node = node ;
                if(bugflag)
                   {
                     printf("Reading node %d\n",l_node) ;
                     fflush(stdout) ;
                   }
                for(i = 0 ; i < dim ; ++i)
                   {
                    fscanf(in_file,"%lf",&value) ;
                    if(l_node)
                      *(array + (l_node-1)*dim + i) = value ;
                   } 
               }
             else
               {
                  for ( i = 0 ; i < gen_order ; ++i)
                    {
                       for( j = 0 ; j < dim ; ++j)
                           fscanf(in_file,"%lf",&temp[j][i]) ;
                    }
                  for ( i = 0 ; i < 3 ; ++i)
                    {
                     fscanf(in_file,"%d",&nodes_gen[i]) ;  
                     fscanf(in_file,"%d",&node_inc[i]) ;  
                    }
                  gen_isopar_grid(array,dim) ;
                  
               }
            }
    }

/************************** end of gen_real **********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_isopar_grid   ***************
*/

/*  INTERFACE:   */

 gen_isopar_grid(
                 real      *array ,  /* array of generated values */
                 int        dim      /* dimensionality of grid */
                )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_isopar_grid
** isopar_grid generates real data via isoparametric interpolation
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int      option = 3 ; /* set to gen over a volume */
      int       incr = 1 ;
      real dr  , ds  , dt    ,
             r  ,  s  ,  t    ;
      real dot_sh ()         ;

      int    l_node ;
      real   value ;
      COUNTER   i , j  ,  k   ,  l  ;

      if(gen_order == 2)
        {
         option = 1 ;       /*  set to gen over a line   */
         if(nodes_gen[0] != 0) incr = node_inc[0] + 1 ;
         if(nodes_gen[1] != 0) incr = node_inc[1] + 1 ;
         if(nodes_gen[2] != 0) incr = node_inc[2] + 1 ;
        }
      else if(gen_order == 4)
             {
              option = 2 ;     /*  set to gen over an area   */
              incr = 1   ;  /* this needs to be made more general */
             }
      else if(gen_order != 8) attempt = GEN_ERR ;


      dr = (nodes_gen[0] != 0) ?   TWO/ (real) nodes_gen[0] : ZERO ;
      ds = (nodes_gen[1] != 0) ?   TWO/ (real) nodes_gen[1] : ZERO ;
      dt = (nodes_gen[2] != 0) ?   TWO/ (real) nodes_gen[2] : ZERO ;

      nodes_gen[0]++ ;
      nodes_gen[1]++ ;
      nodes_gen[2]++ ;

      t = - ONE ;
      for ( k = 0 ; k < *(nodes_gen + 2) ; ++k )
          {
            s = -ONE ;
            for ( j = 0 ; j < *(nodes_gen + 1) ; ++j)
             {
              r = -ONE ;
              for (i = 0 ; i < *nodes_gen ; ++i)
               {

                gen_shape_grid(r,s,t,shape,option) ;
                l_node = node ;
                  for (l = 0 ; l < dim ; ++l)
                     {
                      value = dot_sh(shape,temp[l],gen_order);
                      if(l_node)
                        *(array + (l_node-1)*dim + l) = value ;
                     } 
                node += incr            ;
                r += dr                 ;

               }
              s += ds  ;
             }
           t += dt ;
          }
      }

/************************** end of gen_isopar_grid ***************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_shape_grid   ***************
*/

/*  INTERFACE:   */

  gen_shape_grid(
                 real      r  ,  /* isopar coordinate */
                 real      s  ,  /* isopar coordinate */
                 real      t  ,  /* isopar coordinate */
                 real      *shape  ,  /* shape function storage */
                 int       option      /* generation option flag */
                )

/*  RETURN VALUE:   -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_shape_grid
** gen_shape_grid finds shape function for isoparametric generation.
** It appears to JWP to use bilinear and trilinear interpolation.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      real r1   , r2   , s1   ,
            s2   , t1   , t2   ;
      real *shape1 = shape + 1 ;

      if(option == 1)
        {
         *shape1 = r*PT5 ;
         *shape     = PT5 - *shape1 ;
         *shape1 = PT5 + *shape1 ;
        }
      else if(option == 2)
           {
            r2 = r*PT5 ;
            r1 = PT5-r2 ;
            r2 += PT5   ;
            s2 = s*PT5  ;
            s1 = PT5-s2 ;
            s2 += PT5   ;
            *shape++ = r1*s1 ;
            *shape++ = r2*s1 ;
            *shape++ = r2*s2 ;
            *shape   = r1*s2 ;
           }
           else
           {
            r2 = r*PT5 ;
            r1 = PT5-r2 ;
            r2 += PT5   ;

            s2 = s*PT5  ;
            s1 = PT5-s2 ;
            s2 += PT5   ;

            t2 = t*PT5  ;
            t1 = PT5-t2 ;
            t2 += PT5   ;

            *shape++ = r1*s1*t1 ;
            *shape++ = r2*s1*t1 ;
            *shape++ = r2*s2*t1 ;
            *shape++ = r1*s2*t1 ;
            *shape++ = r1*s1*t2 ;
            *shape++ = r2*s1*t2 ;
            *shape++ = r2*s2*t2 ;
            *shape   = r1*s2*t2 ;
            }
     }

/************************** end of gen_shape_grid ****************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   dot_sh   ***************
*/

/*  INTERFACE:   */

    real  dot_sh(
                 real      *shape  ,     /* shape function storage */
                 real      *variable  ,  /* interpolate variable storage */
                 int        n            /* number to sum over */
                )

/*  RETURN VALUE:   dot product value */
/*  DESCRIPTION:   */
/*
** Function dot_sh
** dot_sh performs the dot product of the vector of shape functions
** (may be gradients of shape functions)
** with the vector "variable"
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

      {
       real sum = ZERO             ,
             *shape_end = shape + n ;

       while (shape < shape_end) sum += (*shape++)*(*variable++) ;

       return(sum) ;
      }

/************************** end of dot_sh ************************************/ 

