/*
***                           File inphase.c                      ***
***                        GeoFEST version 4.3                    ***
*** Copyright (c) 2002, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This software is designated for public release under JPL Task ***
*** Plan Number 83-6791 and may be publicly released after        ***
*** approval for release has been granted by GSFC Software        ***
*** Release  Authority
***
*** This file contains subroutines which handle input, 
*** equation numbering, and memory allocation.
***
***          == input_phase ==
***          == matrix_alloc ==
***          == gen_number ==
***          == output_phase ==
***          == el_output ==
***          == locate_pt ==
***/

/*
***GeoFEST version 4.3***
G. A. Lyzenga -- split node enhancements 8/16/02
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining 
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
4.3: Corrects inefficiency in split node algorithm and introduces new
     split node input format
 $Log: inphase.c,v $
 Revision 1.3  2002/08/23 18:46:49  jwp
 Restored keyword strings, Log and REVNO festures.
 Makefile now has .h dependencies.

 Revision 1.1  2002/07/12 19:05:14  jwp
 Initial revision

*/

#include <stdio.h>
#include "mydefs.h"
#include "errcodes.h"
#define EXTERN extern
#include "finel.h"

static real
   sh_temp[256] , det_temp[9] ;

   char  line[80] ;

real   dotsh() ;

/**********************************************************************/

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   input_phase   ***************
*/

/*  INTERFACE:   */

         input_phase()

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine input_phase
** input_phase is the top-level input file reading function.  
** Most input-file records are read here or in routines 
** in generat.c that are called by this function.
** A few items are read by main.c as well.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     char CVS_Id[] = "@(#) CVS id $Id: inphase.c,v 1.3 2002/08/23 18:46:49 jwp Exp $, Revision: " REVNO;
     COUNTER  i , n ;
     int   numnp , ndof , nrates , nsd ;
     
/* *************************** GLOBAL INPUT PHASE ***************** */

     fscanf(in_file,"%d",&fe_sys.numnp) ;
     fscanf(in_file,"%d",&fe_sys.nsd) ;
     fscanf(in_file,"%d",&fe_sys.ndof) ;
     fscanf(in_file,"%d",&fe_sys.nrates) ;
     fscanf(in_file,"%d",&fe_sys.save_shape) ;
     fscanf(in_file,"%d",&fe_sys.solver) ;

     printf("Number of nodes = %d\n\n",fe_sys.numnp) ;
     printf(
     "nsd = %d  ;  ndof = %d  ;  nrates = %d  ;  save_shape = %d  ;  solver = %d\n\n",
      fe_sys.nsd , fe_sys.ndof , fe_sys.nrates , fe_sys.save_shape , fe_sys.solver ) ;

     numnp = fe_sys.numnp ;
     nsd = fe_sys.nsd ;
     ndof    = fe_sys.ndof    ;
     nrates  = fe_sys.nrates  ;
     
     global.coords = (real * ) calloc(numnp*nsd,sizeof(real)) ;
     if (global.coords == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.displ = (real * ) calloc(numnp*ndof,sizeof(real)) ;
     if (global.displ == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.del_displ = (real * ) calloc(numnp*ndof,sizeof(real)) ;
     if (global.del_displ == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.forv = (real * ) calloc(numnp*ndof,sizeof(real)) ;
     if (global.forv == NULL) attempt = CRD_MEM ;
     completion() ;
     
     global.eqtype = (int * ) calloc(numnp*ndof,sizeof(int)) ;
     if (global.eqtype == NULL) attempt = CRD_MEM ;
     completion() ;
     
     if( nrates > 0 )
        {
         global.bctime = (real * ) calloc(nrates,sizeof(real)) ;
         global.rate = (real * ) calloc(numnp*ndof*nrates,
                                         sizeof(real)) ;
         if (global.rate == NULL) attempt = CRD_MEM ;
         completion() ;
        }
     
/* smoothing is disabled --
     if( fe_sys.psmooth > 0 )
        {
         global.pnode = (real * ) calloc(numnp,sizeof(real)) ;
         global.yp = (real * ) calloc(numnp,sizeof(real)) ;
         if (global.yp == NULL) attempt = PSM_MEM ;
         completion() ;
        }
*/

     gen_map() ;  /*  fill the decomp map array  */
     printf("Global b.c. map generated...\n") ;
     fflush(stdout) ;
         
     gen_number() ;  /* obtain local node and equation numbering */
        
     printf("Equation numbering complete...\n") ; 
     fflush(stdout) ;
    
     gen_real( global.coords,nsd ) ;  /*  fill the coord array  */
     printf("Coordinate generation complete...\n") ;
     fflush(stdout) ;
     
     gen_real( global.forv,ndof ) ; /*  fill the b. c. array  */
     printf("Displacement boundary value generation complete...\n") ;
     fflush(stdout) ;
     
     for( n=0 ; n<nrates ; n++ )
        {
         fscanf(in_file,"%lf",global.bctime+n) ;
         printf("Reading in boundary velocity set #%d for time = %g...\n",
                n+1 , *(global.bctime+n) ) ;
         fflush(stdout) ;
         gen_real( (global.rate + n*ndof*numnp),ndof ) ;
         /*  fill the velocity arrays  */
        }
     printf("Velocity boundary rates generation complete...\n") ;
     fflush(stdout) ;
     
/* ************************ ELEMENT INPUT PHASE ***************** */

     fscanf(in_file,"%d",&global.n_group) ;
     if(global.n_group > MAX_GROUP)
         {
          printf("Too many element groups\n") ;
          exit(1) ;
         }

     for(i=0 ; i<global.n_group ; i++)
         {
          groups[i].el_info =
             (ELEMENT_INFO *) calloc(1,sizeof(ELEMENT_INFO)) ;
          if (groups[i].el_info == NULL) attempt = ELI_MEM + i ;
          completion() ;

          groups[i].el_mat =
             (ELEMENT_MAT *) calloc(1,sizeof(ELEMENT_MAT)) ;
          if (groups[i].el_mat == NULL) attempt = MAT_MEM + i ;
          completion() ;
          
          groups[i].group_num = i+1 ;
         }
                  
     printf("Starting element generation...\n") ;
     fflush(stdout) ;

     elgrp_loop( GENERATE ) ;   /* read in and generate element geometry */

     printf("Doing shape functions...\n") ;
     fflush(stdout) ;

     elgrp_loop( SHAPE ) ;    /* calculate element shape functions
                                 for future use */

  printf("Element generation complete.  Computing column heights.\n") ;
     fflush(stdout) ;
     
/* ************************ EQUATION SETUP PHASE ***************** */

     matrix_alloc() ;      /*  set up storage and indexing for the 
                              skyline profile elimination solver OR
                              the preconditioned CG solver  */


     printf("Equation setup completed.\n") ;
     fflush(stdout) ;

/* ************************ TIME INPUT PHASE ***************** */

     time_data.time = ZERO ;
     
     fscanf(in_file,"%d%d",&time_data.nprt_nodes,&time_data.nprt_elem) ;

     if(time_data.nprt_nodes > 0)
        {
         time_data.node_list
            = ( int  * ) calloc(time_data.nprt_nodes,sizeof(int)) ;
         if (time_data.node_list == NULL) attempt = TGP_MEM ;
         completion() ;

         for( i = 0 ; i < time_data.nprt_nodes ; i++ )
            fscanf(in_file,"%d", time_data.node_list + i ) ;
        }

     if(time_data.nprt_elem > 0)
        {
         time_data.elem_list
            = ( int  * ) calloc(2*time_data.nprt_elem,sizeof(int)) ;
         if (time_data.elem_list == NULL) attempt = TGP_MEM ;
         completion() ;

         for( i = 0 ; i < time_data.nprt_elem ; i++ )
            fscanf(in_file,"%d%d", time_data.elem_list + 2*i
                   , time_data.elem_list + 2*i + 1 ) ;
                                    /* group # and elem. # */
        }

     fscanf(in_file,"%d%d%d%lf",&time_data.ntime_grp,&time_data.nreform,
                           &time_data.nbackup,&time_data.slip_interval) ;
     
     if(time_data.ntime_grp == 0)  /* this is an elastic-only run */
        {
         time_data.nprints = 1 ;
         time_data.prt_time
             = ( real  * ) calloc(time_data.nprints,sizeof(real)) ;
         *(time_data.prt_time) = ZERO ;
         return(0) ;
        }
     
     /* the rest of the input required only for time-stepping runs... */
     
     time_data.maxstep
        = ( real  * ) calloc(time_data.ntime_grp,sizeof(real)) ;
     if (time_data.maxstep == NULL) attempt = TGP_MEM ;
     time_data.endtime
        = ( real  * ) calloc(time_data.ntime_grp,sizeof(real)) ;
     if (time_data.endtime == NULL) attempt = TGP_MEM ;
     time_data.nsteps
        = ( int  * ) calloc(time_data.ntime_grp,sizeof(int)) ;
     if (time_data.nsteps == NULL) attempt = TGP_MEM ;
     completion() ;
                 
     time_data.alpha
         = ( real  * ) calloc(time_data.ntime_grp,sizeof(real)) ;
     if (time_data.alpha == NULL) attempt = TGP_MEM ;
     completion() ;
                 
     time_data.delt 
         = ( real  * ) calloc(time_data.ntime_grp,sizeof(real)) ;
     if (time_data.delt == NULL) attempt = TGP_MEM ;
     completion() ;
                 
     for( i = 0 ; i < time_data.ntime_grp ; i++ )
        fscanf(in_file,"%lf%lf%lf",
             time_data.endtime + i, time_data.alpha + i , time_data.delt + i ) ;

     fscanf(in_file,"%d", &time_data.nprints ) ;
     if( time_data.nprints > 0 )
         {
          time_data.prt_time
             = ( real  * ) calloc(time_data.nprints,sizeof(real)) ;
          if (time_data.prt_time == NULL) attempt = TGP_MEM ;
          completion() ;

          for( i = 0 ; i < time_data.nprints ; i++ )
            fscanf(in_file,"%lf", time_data.prt_time + i ) ;
         }

     printf("Time data  generation complete...\n") ;
     fflush(stdout) ;
     
     time_data.restart = 0 ;
     fscanf(in_file,"%s",line) ;
     if(strcmp(line , "NO_RESTART") != 0 )
        {
         sprintf(time_data.startfile,"%s",line) ;
         time_data.restart = 1 ;
         vrestart() ;
        }

     time_data.sav_state = 0 ;
     fscanf(in_file,"%s",line) ;
     if(strcmp(line , "NO_SAVE") != 0 )
        {
         sprintf(time_data.savefile,"%s",line) ;
         time_data.sav_state = 1 ;
        }

    } /* end function */

/************************** end of input_phase *******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   matrix_alloc   ***************
*/

/*  INTERFACE:   */

         matrix_alloc()

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine matrix_alloc
** matrix allocates space for the sparse stiffness matrix _or_
** the PCG solver arrays
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int   i ;
    
    switch(fe_sys.solver)
       {
        case DIRECT:
         stiff.neq = fe_sys.neq ;

         stiff.diag = (int  * ) calloc(fe_sys.neq,sizeof(int )) ;
         if (stiff.diag == NULL) attempt = DIAG_MEM ;
         completion() ;
         
         elgrp_loop( REORDER );
         elgrp_loop( COL_HT ) ;
         profile_diag( &stiff , INTERIOR ) ;
         
         printf("Total profile size is %d\n",stiff.size);

         stiff.matrix =
            (real * ) calloc(stiff.size,sizeof(real)) ;
         if (stiff.matrix == NULL) attempt = PROF_MEM ;
         completion() ;

         break;
/*---------------------------------------------------------------*/
         
        case PCG:
         pcg.neq = fe_sys.neq ;

         pcg.precond = (real  * ) calloc(fe_sys.neq,sizeof(real)) ;
         if (pcg.precond == NULL) attempt = PCG_MEM ;

         pcg.r = (real  * ) calloc(fe_sys.neq,sizeof(real)) ;
         if (pcg.r == NULL) attempt = PCG_MEM ;

         pcg.d = (real  * ) calloc(fe_sys.neq,sizeof(real)) ;
         if (pcg.d == NULL) attempt = PCG_MEM ;

         pcg.temp = (real  * ) calloc(fe_sys.neq,sizeof(real)) ;
         if (pcg.temp == NULL) attempt = PCG_MEM ;

         completion() ;
         
         pcg.hist = (real  * ) calloc(200,sizeof(real)) ;
         
         if((cgconv_file = fopen("cghist.txt","w")) == NULL)
          {
           printf("Convergence history file couldn't be opened!\n") ;
           exit(1) ;
          }

         global.permut = (int *) calloc(fe_sys.numnp+1,sizeof(int));
         for(i=0;i<=fe_sys.numnp;i++)  global.permut[i] = i ;  /* null permutation */

         force.last_result = (real * ) calloc(fe_sys.neq,sizeof(real)) ;
         if (force.last_result == NULL) attempt = R_MEM ;

         break;
       }
       
         force.full_rhs = (real * ) calloc(fe_sys.neq,sizeof(real)) ;
         if (force.full_rhs == NULL) attempt = R_MEM ;
         completion() ;

   }

/************************** end of matrix_alloc ******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   gen_number   ***************
*/

/*  INTERFACE:   */

         gen_number()

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine gen_number
** gen_number assigns equation numbers based on nodes and ndof
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    COUNTER   i , j , k , index ;
    int   eqtype[3] ;
    int   ieq_local ;
    int   *id_pointer ;
    int   neq_local , ndof , numnp ;

     neq_local = 0 ;
     numnp = fe_sys.numnp ;
     ndof = fe_sys.ndof ; 

   
     for( i=0 ; i< numnp ; i++ )
         {
          index = i ;

          for(k=0 ; k<ndof ; k++)
             {
              eqtype[k] = *(global.eqtype+ndof*index+k) ;
             }
                      /* 1 = active , 0 = boundary  */

               for(k=0 ; k<ndof ; k++)
                  if(eqtype[k])  neq_local++ ;
              
         }   /* end loop over global node number */

     fe_sys.neq = neq_local ;
     printf ("numnp = %d  neq = %d\n", fe_sys.numnp, fe_sys.neq) ;
     fflush(stdout) ;
     
  /* we now will assign equation numbers */
  
  /* first, allocate space for the id array  */
     global.id_pointer =
        (int * ) calloc(numnp*ndof,sizeof(int)) ;
     if (global.id_pointer == NULL) attempt = ID_MEM ;
     completion() ;

     id_pointer = global.id_pointer ;
     
  /* now assign equations ...    (equations start at zero)  */
   
     ieq_local = 0 ;
     
     for(j=0 ; j<numnp*ndof ; j++) *(id_pointer + j) = (-1) ;
     
     for( i=0 ; i<numnp ; i++ )
         {
          index = i ;

          for(k=0 ; k<ndof ; k++)
             {
              eqtype[k] = *(global.eqtype+ndof*index+k) ;
             }

          for(k=0 ; k<ndof ; k++)
             {
              if(eqtype[k])
                 {
                  *(id_pointer + ndof*(index) + k) =
                     ieq_local ;
                  ieq_local++ ;
                 }
             }
         }   /* end loop over global node number */
         

   }

/************************** end of gen_number ********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   output_phase   ***************
*/

/*  INTERFACE:   */

   output_phase(
                int    quake   /* flag to indicate quake or not */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine output_phase
** output_phase writes out requested data at a quake 
** or scheduled time.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     COUNTER  i , n , k , ntot , index , ndof , numnp , nsd ;
     int  mine , print , curr_node ;
     static real  outbuf[9] ;
     real  time , test ;
         
     ndof = fe_sys.ndof ;
     numnp = fe_sys.numnp ;
     nsd = fe_sys.nsd ;
     time = time_data.time ;
     
     print = 0 ;
     for(n=0 ; n<time_data.nprints ; n++)
        {
         test = *(time_data.prt_time + n) ;
         if( test >= ZERO && test <= time )
            {
             *(time_data.prt_time + n) = (-ONE) ;
             print = 1 ;
             break ;
            }
        }
     if( !(print || quake) )
         return(0) ;

     if(quake)
       printf("Performing quake output at time = %f ...\n",time) ;
     else
       printf("Performing scheduled output at time = %f ...\n",time) ;

     fprintf(out_file,
     "\n Global coordinates & displacements & delt displacements \n") ;
     fprintf(out_file,"\n Simulation time = %f ; step size = %g\n",
             time,time_data.dt) ;
     if(quake)
        fprintf(out_file,"***Fault Rupture*** at t=%g\n",time) ;
     fprintf(out_file,
 "\n          coordinates          displacements       del_displacements\n") ;
     
             if(time_data.nprt_nodes == (-1)) ntot = numnp ;
             else  ntot = time_data.nprt_nodes ;
             
             for(i=0 ; i<ntot ; i++)
                 {
                  mine = 0 ;
                  if(time_data.nprt_nodes == (-1))
                     curr_node = i+1 ;
                  else
                     curr_node = *(time_data.node_list+i) ;

                  index = curr_node - 1 ;
                  if( index != (-1) ) mine = 1 ;
        
                  if(mine)
                     {
                      for(k=0; k<nsd; k++)
                          outbuf[k] = *(global.coords + nsd*index + k) ;

                      for(k=0 ; k<ndof ; k++)
                         {
                          outbuf[nsd+k] = *(global.displ + ndof*index + k) ;
                          outbuf[nsd+ndof+k] = *(global.del_displ +
                                                         ndof*index + k) ;
                         }
                     }

                  fprintf(out_file,"node %d",curr_node);

                  for(k=0; k<nsd; k++)
                      fprintf(out_file,"   %f", outbuf[k]) ;

                  for(k=0 ; k<2*ndof ; k++)
                     {
                      fprintf(out_file,"   %g",outbuf[nsd+k]) ;
                     }
                  fprintf(out_file,"\n") ;
        
                 }        
             fflush(out_file) ;
             elgrp_loop(OUTPUT) ;
             printf("Output step finished.\n") ;
    }

/************************** end of output_phase ******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   el_output   ***************
*/

/*  INTERFACE:   */

      el_output(
                GROUP   *grp_ptr    /* pointer to current element group */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine el_output
** el_output prints out element stress information.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int     numel , nint , nel , use_list , n , str_size ,
              ipt , j , ng , ntot , nsd ;
      real   xpt[3][10] , str_ctr[6] ;
      ELEMENT_DATA   *el_pt ;


     fprintf(out_file,
     "\n Element stresses  -- element group #%d \n",grp_ptr->group_num) ;
     fprintf(out_file,"\n Simulation time = %f ; step size = %g\n",
             time_data.time,time_data.dt) ;
     fprintf(out_file,
 "\n coordinates                          stresses . . .->      el.#\n") ;

      fflush(out_file) ;
      nsd = fe_sys.nsd ;
      numel = (grp_ptr->el_info)->numel ;
      
      str_size = (grp_ptr->el_info)->stress_size ;
      
      switch((grp_ptr->el_info)->type)
          {
           case BILIN:
           case SERENDIP:
           case TET:
           case TRILIN:
               if(time_data.nprt_elem == (-1))
                  {
                   ntot = (grp_ptr->el_info)->numel ;
                   use_list = 0 ;
                  }
               else
                  {
                   ntot = time_data.nprt_elem ;
                   use_list = 1 ;
                  }
               for ( nel = 0 ; nel < ntot ; nel++)
                   {
                    fflush(out_file) ;
                    if( use_list )
                       {
                        if( *(time_data.elem_list + 2*nel) !=
                            grp_ptr->group_num ) continue ;
                        ng = *(time_data.elem_list + 2*nel + 1) ;
                        n = ng ;
                       }
                    else
                       {
                        ng = nel+1 ;
                        n = ng ;
                       }
                    if( n == 0 ) continue ;
                    
                    el_pt = grp_ptr->el_data + n - 1 ;
                    nint = el_pt->nint ;
                    locate_pt( el_pt , grp_ptr->el_info , xpt ) ;

                    for(j=0 ; j<str_size ; j++)
                       str_ctr[j] = ZERO ;  /* midpoint stresses */

                    for(j=0 ; j < nsd; j++)
                       xpt[j][9] = ZERO ;  /* midpoint location */
                   

                    for (ipt = 0 ; ipt < nint ; ipt++)
                       {
                        for(j=0; j < nsd; j++)
                           xpt[j][9] += xpt[j][ipt] / ((real) nint) ;

                        for(j=0 ; j<str_size ; j++)
                           str_ctr[j] +=
                 *((el_pt -> stress) + ipt*str_size + j)  / ((real) nint) ;
                       }

                    for(j=0 ; j < nsd; j++)
                       fprintf(out_file,"  %f  ", xpt[j][9] ) ;
                    for(j=0 ; j<str_size ; j++)
                       fprintf( out_file,"%f     ", str_ctr[j] ) ;
                    fprintf(out_file,"   %d\n" , ng) ;
                   }
               fflush(out_file) ;
               break ;
               
               
          }

    }

/************************** end of el_output *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   locate_pt   ***************
*/

/*  INTERFACE:   */

      locate_pt(
                ELEMENT_DATA  *el_pt  ,  /* ptr to current element struct */
                ELEMENT_INFO  *info ,   /* ptr to element info struct */
                real   xpt[3][10]     /* return array of computed coords */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine locate_pt
** locate_pt maps the integration points into physical space.
** Global coordinates are placed in xpt[j]
** Called by el_output element stress output function (above).
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

  {
   int    j , ipt , save_shape , type , nint , nen , nsd , nsdp1 ;
   real  *shape_ptr , *det_ptr , *sh ;

   save_shape = fe_sys.save_shape ;
   nen = info->nen ;
   type = info->type ;
   nint = el_pt->nint ;
   nsd = fe_sys.nsd ;
   nsdp1 = nsd + 1 ;
   
   if( !save_shape )
        {
         shape_ptr = sh_temp ;
         det_ptr = det_temp ;
         shape( el_pt , info , shape_ptr , det_ptr , type ,
                nen , nint , 0 ) ;
        }
   else
        {
         shape_ptr = el_pt -> shape ;
        }

     for (ipt = 0 ; ipt < nint ; ipt++)
          {
           sh = shape_ptr + ipt*nen*nsdp1 ;

           for(j=0; j < nsd ; j++)
              xpt[j][ipt] = dotsh( global.coords ,
                  nsd , j , sh , el_pt , info , nen , nsd , 0 ) ;
          }

  }

/************************** end of locate_pt *********************************/ 
